package org.gersteinlab.regulation;

import org.gersteinlab.regulation.util.*;

import java.io.*;
import java.util.*;

/**
 *	<pre>
 *	A class for analyzing the prediction results.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (October 15, 2008)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class Analyzer
{
	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	public static void main(String[] argv)
	{
		try
		{
			// Store parameter values
			int paramCount = 0;
			File ansFile = new File(argv[paramCount++]);
			File logFile = new File(argv[paramCount++]);
			File preFile = new File(argv[paramCount++]);
			File anaFile = new File(argv[paramCount++]);

			// Read the gold standard regulations
			Set<String> goldInts = new HashSet<String>();
			{
				BufferedReader br = new BufferedReader(new FileReader(ansFile));
				String line = br.readLine();
				while (line != null && line.length() != 0)
				{
					StringTokenizer2 st = new StringTokenizer2(line);
					String protein1 = st.nextToken();
					String protein2 = st.nextToken();
					boolean interact = st.nextToken().equals("1");
					if (interact)
						goldInts.add(protein1 + "\t" + protein2);
					line = br.readLine();
				}
				br.close();
			}

			// Get the number of predictions of each batch
			int[] batchPredCounts = new int[7];
			{
				BufferedReader br = new BufferedReader(new FileReader(logFile));
				String line = br.readLine();
				for (int i=0; i<7; i++)
				{
					while (line.indexOf("Number of predictions in") == -1)
						line = br.readLine();
					batchPredCounts[i] = Integer.parseInt(line.substring(line.indexOf(":")+2));
					line = br.readLine();
				}
				br.close();
			}

			// Count the number of correct predictions in each batch
			List<Prediction> predictions = new ArrayList<Prediction>();
			int posCount = 0;
			int[] correctPredCounts = new int[7];
			{
				BufferedReader br = new BufferedReader(new FileReader(preFile));
				PrintWriter pw =  new PrintWriter(new FileWriter(anaFile));
				pw.println("Batch\tRegulator\tTarget\tPred. prob.\tTrue class");
				for (int i=0; i<7; i++)
				{
					for (int j=0; j<batchPredCounts[i]; j++)
					{
						String line = br.readLine();
						StringTokenizer2 st = new StringTokenizer2(line);
						String protein1 = st.nextToken();
						String protein2 = st.nextToken();
						int trueClass = 0;
						if (goldInts.contains(protein1 + "\t" + protein2))
						{
							correctPredCounts[i] ++;
							trueClass = 1;
						}

						double predProb = Double.parseDouble(st.nextToken());
						pw.println((i + 1) + "\t" + protein1 + "\t" + protein2 + "\t" + predProb + "\t" + trueClass);

						predictions.add(new Prediction(protein1, protein2, trueClass, predProb));
						if (trueClass == 1)
							posCount ++;
					}
				}
				br.close();
				pw.close();
			}

			// Output results
			System.out.println(preFile.getName());
			for (int i=0; i<7; i++)
				System.out.println(batchPredCounts[i] + "\t" + correctPredCounts[i]);
			System.out.println("AUC: " + getAUC(predictions, posCount));
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}

	/**
	 *	Get the area under curve (AUC) of a receiver-operator
	 *	characteristics (ROC) curve, for the whole set of predictions
	 *	and at some specific reporting points.
	 *	@param		predictions		The predictions made
	 *	@param		posCount		The total number of real positive examples
	 */
	protected static double getAUC(List<Prediction> predictions, int posCount)
	{
		// Sort the predictions in descending order of the probability
		// of being in the positive class
		Collections.sort(predictions);

		double aucTemp		= 0;	// Total un-normalized area up to now
		int tpCount		= 0;	// Number of positive examples encountered up to now
		int fpCount		= 0;	// Number of negative examples encountered up to now

		// Variables for handling ties
		double prevPredProb	= -1;	// The probability of the previous prediction
		int bufferedTPCount	= 0;	// The number of positive examples encoutnered, but not yet recorded
		int bufferedFPCount	= 0;	// The number of negative examples encoutnered, but not yet recorded

		// Scan the sorted list to compute the AUC values
		for (Prediction prediction : predictions)
		{
			if (prediction.predProb != prevPredProb)	// Not a tie
			{
				aucTemp += bufferedFPCount * (tpCount + 0.5 * bufferedTPCount);
				tpCount += bufferedTPCount;
				fpCount += bufferedFPCount;
				bufferedTPCount = 0;
				bufferedFPCount = 0;
				prevPredProb = prediction.predProb;
			}

			if (prediction.trueClass == 1)
				bufferedTPCount ++;
			else
				bufferedFPCount ++;
		}

		// Clear any buffered values
		aucTemp += bufferedFPCount * (tpCount + 0.5 * bufferedTPCount);
		tpCount += bufferedTPCount;
		fpCount += bufferedFPCount;

		double auc = aucTemp / posCount / fpCount;
		return auc;
	}


	/*----------------------------------------------------------------------
	 Inner classes
	----------------------------------------------------------------------*/

	/**
	 *	This class represents the prediction result for a regulation
	 *	event.
	 */
	protected static class Prediction implements Comparable<Prediction>
	{
		/**
		 *	The regulator
		 */
		String regulator;

		/**
		 *	The target
		 */
		String target;

		/**
		 *	The correct class label of the pair: +1 for a
		 *	regulation, 0 for a non-regulation
		 */
		int trueClass;

		/**
		 *	Predicted probability of regulation
		 */
		double predProb;

		public Prediction(String regulatorIn, String targetIn, int trueClassIn, double predProbIn)
		{
			this.regulator	= regulatorIn;
			this.target	= targetIn;
			this.trueClass	= trueClassIn;
			this.predProb	= predProbIn;
		}

		// Larger predicted probability first
		public int compareTo(Prediction p2)
		{
			     if (this.predProb < p2.predProb)
				return 1;
			else if (this.predProb > p2.predProb)
				return -1;
			else
				return 0;
		}
	}
}