package org.gersteinlab.regulation;

import org.gersteinlab.regulation.util.*;
import JSci.maths.statistics.*;
import java.io.*;

/**
 *	<pre>
 *	A class for storing the deletion profile of all genes read from a file.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (August 20, 2008)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class DeletionProfile
{
	/*----------------------------------------------------------------------
	 Instance variables
	----------------------------------------------------------------------*/

	protected int geneCount = -1;
	protected String[] geneNames = null;
	protected double[][] profile = null;	// First dimension: measured gene
						// Second dimension: deleted gene
						// Wild type moved to the end of each vector

	// Statistics
	protected double eMean = 0;		// Mean error
	protected double eSD = 0;		// Standard deviation of error
	protected double[] vMeans = null;	// Mean of unaffected values


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	/**
	 *	Read and return the profile in a file.
	 *	@param		r			The reader of the file
	 *	@return					The profile
	 */
	public static DeletionProfile readProfile(Reader r) throws IOException
	{
		DeletionProfile dp = new DeletionProfile();
		BufferedReader br = new BufferedReader(r);
		String line = br.readLine();

		// Get the number of genes and create the profile structure
		StringTokenizer2 st = new StringTokenizer2(line);
		dp.geneCount = st.countTokens() - 1;
		dp.geneNames = new String[dp.geneCount];
		st.nextToken();
		for (int i=0; i<dp.geneCount; i++)
			dp.geneNames[i] = st.nextToken();
		dp.profile = new double[dp.geneCount][];
		for (int i=0; i<dp.geneCount; i++)
			dp.profile[i] = new double[dp.geneCount+1];

		// Read the profiles
		for (int i=0; i<=dp.geneCount; i++)
		{
			int iIndex = (i == 0) ?(dp.geneCount) :(i-1);

			line = br.readLine();
			st = new StringTokenizer2(line);
			st.nextToken();		// Skip the header column
			for (int j=0; j<dp.geneCount; j++)
				dp.profile[j][iIndex] = Double.parseDouble(st.nextToken());
		}
		br.close();

		// Compute the statistics
		dp.computeStats();

		return dp;
	}

	/**
	 *	1. Initialize the mean of each gene by its wild type value
	 *	2. Identify points not affected by the deletions
	 *	3. Estimate mean and sd of global random error
	 *	4. Use the global random error to re-estimate mean of each gene
	 *	5. Repeat 2-4
	 */
	protected void computeStats()
	{
		// Initialize vMeans
		vMeans = new double[geneCount];
		for (int i=0; i<geneCount; i++)
			vMeans[i] = profile[i][geneCount];

		for (int iter=0; iter<10; iter++)
		{
			// Identify unaffected genes, and estimate global error
			double eCount = 0;
			double eSum = 0;
			double eSumSq = 0;

			for (int i=0; i<geneCount; i++)
			{
				double sdI = SimpleStats.getSD(profile[i]);
				for (int j=0; j<geneCount; j++)
				{
					if (i == j)		// Must be affected by itself
						continue;

					double sdIJ = (profile[i][j] - vMeans[i]) / sdI;
					if (Math.abs(sdIJ) < 1)
					{
						eCount ++;
						eSum += profile[i][j] - vMeans[i];
						eSumSq += (profile[i][j] - vMeans[i]) * (profile[i][j] - vMeans[i]);
					}
				}
			}

			eMean = eSum / eCount;
			eSD = Math.pow(eSumSq / eCount - eMean * eMean, 0.5);
	
			// Re-estimate vMean
			for (int i=0; i<geneCount; i++)
			{
				double vCount = 0;
				double vSum = 0;
				for (int j=0; j<geneCount; j++)
				{
					if (i == j)		// Must be affected by itself
						continue;

					double esdIJ = (profile[i][j] - vMeans[i]) / eSD;
					if (Math.abs(esdIJ) < 2)
					{
						vCount ++;
						vSum += profile[i][j];
					}
				}
				double wtI = profile[i][geneCount];
				if (vCount != 0	&&				// To avoid division by zero
				    Math.abs((wtI - vSum / vCount) / eSD) <= 2)	// To avoid genes affected by many deletions in the same direction
					vMeans[i] = vSum / vCount;
			}
		}
	}

	public double[][] getReactionProbs()
	{
		double[][] reactionProbs = new double[geneCount][];	// Dimension 1: measured gene
									// Dimension 2: deleted gene
		for (int i=0; i<geneCount; i++)
			reactionProbs[i] = new double[geneCount];
		NormalDistribution norm = new NormalDistribution();

		for (int i=0; i<geneCount; i++)
		{
			for (int j=0; j<geneCount; j++)
			{
				if (i == j)
					reactionProbs[i][j] = 0;
				else
				{
					double esdIJ = -1 * Math.abs(profile[i][j] - vMeans[i]) / eSD;
					reactionProbs[i][j] = 1 - norm.cumulative(esdIJ) * 2;
					// Safety check
					if (Double.isNaN(reactionProbs[i][j]))
						throw new RuntimeException("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!: ");
				}
			}
		}
/*
		// Search for potential essential genes
		Outer: for (int j=0; j<geneCount; j++)
		{
			for (int i=0; i<geneCount; i++)
			{
				if (profile[i][j] > 5 * eSD)
					continue Outer;
			}
			System.out.println("Potential essential gene: " + j);
			for (int i=0; i<geneCount; i++)
				reactionProbs[i][j] = Double.NaN;
		}
*/
		return reactionProbs;
	}
}