package org.gersteinlab.regulation;

import org.gersteinlab.regulation.model.*;
import org.gersteinlab.regulation.numeric.*;
import org.gersteinlab.regulation.util.*;

import java.io.*;
import java.util.*;

/**
 *	<pre>
 *	The main predictor class.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (August 20, 2008)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class Main
{
	/*----------------------------------------------------------------------
	 Constants
	----------------------------------------------------------------------*/

	public static final String TEAM_NAME = "BTEAM";
	protected static final int SEARCH_ITER = 100;	// Number of iteractions of local search


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	public static void main(String[] argv)
	{
		try
		{
			// Store parameter values
			File dataBase = new File(argv[0]);
			File resultsBase = new File(argv[1]);
			int geneCount = Integer.parseInt(argv[2]);
			String setID = argv[3];
			int polyDeg = Integer.parseInt(argv[4]);
			System.out.println("Working on Insilico" + setID);

			// Read profiles
			File nullFile = new File(dataBase, "InSilico" + setID + "-null-mutants.tsv");
			File heteFile = new File(dataBase, "InSilico" + setID + "-heterozygous.tsv");
			File trajFile = new File(dataBase, "InSilico" + setID + "-trajectories.tsv");

			FileReader nullReader = new FileReader(nullFile);
			FileReader heteReader = new FileReader(heteFile);
			FileReader trajReader = new FileReader(trajFile);

			DeletionProfile   nullProfile = DeletionProfile.readProfile(nullReader);
			DeletionProfile   heteProfile = DeletionProfile.readProfile(heteReader);
			TrajectoryProfile trajProfile = TrajectoryProfile.readProfile(trajReader, polyDeg);

			nullReader.close();
			heteReader.close();
			trajReader.close();

			// Store the smoothed trajectories
			File smoothedTrajFile = new File(dataBase, "InSilico" + setID + "-trajectories_smoothed" + polyDeg + ".tsv");
			PrintWriter smoothedTrajWriter = new PrintWriter(new FileWriter(smoothedTrajFile));
			smoothedTrajWriter.print("Time");
			for (int i=0; i<trajProfile.geneCount; i++)
				smoothedTrajWriter.print("\tG" + i);

			for (int j=0; j<trajProfile.expCount; j++)
			{
				smoothedTrajWriter.println();
				Polynomial[] ps = new Polynomial[trajProfile.geneCount];
				for (int i=0; i<trajProfile.geneCount; i++)
					ps[i] = new Polynomial(trajProfile.profile2[i][j]);

				for (int k=0; k<trajProfile.pointCount; k++)
				{
					smoothedTrajWriter.print(k * 10);
					for (int i=0; i<trajProfile.geneCount; i++)
						smoothedTrajWriter.print("\t" + ps[i].f(trajProfile.xs[k]));
					smoothedTrajWriter.println();
				}
			}
			smoothedTrajWriter.close();

			// Variables for the final predictions
			List<String> predictions = new ArrayList<String>();	// <regulator>\t<target>
			int[][] predicted = new int[geneCount][];		// Frist dimension: regulator
										// Second dimension: target
			for (int i=0; i<geneCount; i++)
				predicted[i] = new int[geneCount];

			// Get the reaction probabilities from deletion strains
			double[][] nullReacProbs = nullProfile.getReactionProbs();
			double[][] heteReacProbs = heteProfile.getReactionProbs();

			String[] delProfileNames = {"null-mutants", "heterzygous"};
			double[][][] delReacProbs = {nullReacProbs, heteReacProbs};

			for (int d=0; d<delProfileNames.length; d++)
			{
				File probsFile = new File(resultsBase, "InSilico" + setID + "-" + delProfileNames[d] + "_probs.txt");
				PrintWriter probsWriter = new PrintWriter(new FileWriter(probsFile));
				for (int i=0; i<geneCount; i++)
					probsWriter.print("\tGene " + (i + 1));
				probsWriter.println();
				for (int j=0; j<geneCount; j++)
				{
					probsWriter.print("Deleted " + (j + 1));
					for (int i=0; i<geneCount; i++)
						probsWriter.print("\t" + delReacProbs[d][i][j]);
					probsWriter.println();
				}
				probsWriter.close();
			}

			// First batch of predictions: prob > 0.99 from null mutants
			{
				List<int[]> tempPredictions = new ArrayList<int[]>();
				for (int i=0; i<geneCount; i++)
					for (int j=0; j<geneCount; j++)
						if (nullReacProbs[i][j] > 0.99)
							tempPredictions.add(new int[] {j, i});
				double[] probs = new double[tempPredictions.size()];
				int[][] pairs = new int[tempPredictions.size()][];
				for (int i=0; i<tempPredictions.size(); i++)
				{
					pairs[i] = tempPredictions.get(i);
					probs[i] = nullReacProbs[pairs[i][1]][pairs[i][0]];
				}
				Util.coSort(probs, pairs, false);

				int newPredictions = 0;
				for (int i=0; i<pairs.length; i++)
				{
					predictions.add("G" + (pairs[i][0] + 1) + "\tG" + (pairs[i][1] + 1));
					predicted[pairs[i][0]][pairs[i][1]] = (nullProfile.profile[pairs[i][1]][pairs[i][0]] < nullProfile.vMeans[pairs[i][1]]) ?1 :-1;
					newPredictions ++;
				}
				System.out.println("Number of predictions in batch 1: " + newPredictions);
			}

			// Get the residuals from perturbation trajectories
			// First dimension: target gene
			// Second dimension: potential regulator
			double[][] linRes = new double[geneCount][];
			double[][] sigRes = new double[geneCount][];
			double[][] mulRes = new double[geneCount][];

//			String[] modelNames = {"linear", "sigmoidal", "multiplicative"};
			String[] modelNames = {"linear", "sigmoidal"};
//			double[][][] residuals = {linRes, sigRes, mulRes};
			double[][][] residuals = {linRes, sigRes};

			for (int m=0; m<modelNames.length; m++)
			{
				for (int i=0; i<geneCount; i++)
					residuals[m][i] = new double[geneCount];

				File residualFile = new File(resultsBase, "InSilico" + setID + "-trajectories_" + modelNames[m] + "-residual.txt");
if (residualFile.exists())
{
	BufferedReader residualReader = new BufferedReader(new FileReader(residualFile));
	String line = residualReader.readLine();
	for (int i=0; i<geneCount; i++)
	{
		line = residualReader.readLine();
		StringTokenizer st = new StringTokenizer(line, "\t", false);
		st.nextToken();
		for (int j=0; j<geneCount; j++)
			residuals[m][j][i] = Double.parseDouble(st.nextToken());
	}
	residualReader.close();
}
else
{
				PrintWriter residualWriter = new PrintWriter(new FileWriter(residualFile));
				residualWriter.print(modelNames[m]);
				for (int i=0; i<geneCount; i++)
					residualWriter.print("\tGene " + (i + 1));
				residualWriter.println();
				for (int j=0; j<geneCount; j++)
				{
					residualWriter.print("Regulator " + (j + 1));
					for (int i=0; i<geneCount; i++)
					{
						if (j == i)
							residualWriter.print("\t" + Double.NaN);
						else
						{
							if (modelNames[m].equals("linear"))
							{
								int[][] preds = {{j}};
								double[] coefs = new double[3];
								double residual = trajProfile.getResidual(i, preds, LinearRateModel.class, coefs, SEARCH_ITER, new double[] {-1, -1, -1}, new double[] {1, 1, 1});
								residuals[m][i][j] = (coefs[2] < 0) ?-residual :residual;
								residualWriter.print("\t" + residuals[m][i][j]);
							}
							else if (modelNames[m].equals("sigmoidal"))
							{
								int[][] preds = {{j}};
								double[] coefs = new double[4];
								double residual = trajProfile.getResidual(i, preds, SigmoidalRateModel.class, coefs, SEARCH_ITER, new double[] {-1, -1, -1, -1}, new double[] {1, 1, 1, 1});
								residuals[m][i][j] = (coefs[3] > 0) ?-residual :residual;
								residualWriter.print("\t" + residuals[m][i][j]);
							}
							else if (modelNames[m].equals("multiplicative"))
							{
								// First as inhibitor
								int[][] predsI = {{j},{}};
								double[] coefsI = new double[4];
								double residualI = trajProfile.getResidual(i, predsI, MultiplicativeRateModel.class, coefsI, SEARCH_ITER, new double[] {0, 0, 0, 0}, new double[] {1, 1, 1, 1});

								// Then as enhancer
								int[][] predsE = {{}, {j}};
								double[] coefsE = new double[4];
								double residualE = trajProfile.getResidual(i, predsE, MultiplicativeRateModel.class, coefsE, SEARCH_ITER, new double[] {0, 0, 0, 0}, new double[] {1, 1, 1, 1});

								residuals[m][i][j] = (residualI < residualE) ?-residualI :residualE;
								residualWriter.print("\t" + residuals[m][i][j]);
							}
						}
					}
					residualWriter.println();
				}
				residualWriter.close();
}
			}

			// Second batch of predictions: residual < 1.5 SD from
			// mean among all potential regulators of a gene, as
			// predicted by all models
			predictIntersection(residuals, predictions, predicted, 2);

			// Get the residuals of guided models using the initial
			// guesses in batch 1 and batch 2
			// First dimension: target gene
			// Second dimension: potential regulator
			double[][] linGuidedRes = new double[geneCount][];
			double[][] sigGuidedRes = new double[geneCount][];
			double[][] mulGuidedRes = new double[geneCount][];

//			double[][][] guidedResiduals = {linGuidedRes, sigGuidedRes, mulGuidedRes};
			double[][][] guidedResiduals = {linGuidedRes, sigGuidedRes};

			for (int m=0; m<modelNames.length; m++)
			{
				for (int i=0; i<geneCount; i++)
				{
					guidedResiduals[m][i] = new double[geneCount];
					Arrays.fill(guidedResiduals[m][i], Double.NaN);
				}

				File guidedResidualFile = new File(resultsBase, "InSilico" + setID + "-trajectories_" + modelNames[m] + "-guided-residual.txt");
if (guidedResidualFile.exists())
{
	BufferedReader guidedResidualReader = new BufferedReader(new FileReader(guidedResidualFile));
	String line = guidedResidualReader.readLine();
	for (int i=0; i<geneCount; i++)
	{
		line = guidedResidualReader.readLine();
		StringTokenizer st = new StringTokenizer(line, "\t", false);
		st.nextToken();
		for (int j=0; j<geneCount; j++)
			guidedResiduals[m][j][i] = Double.parseDouble(st.nextToken());
	}
	guidedResidualReader.close();
}
else
{
				PrintWriter guidedResidualWriter = new PrintWriter(new FileWriter(guidedResidualFile));
				guidedResidualWriter.print("guided " + modelNames[m]);
				for (int i=0; i<geneCount; i++)
					guidedResidualWriter.print("\tGene " + (i + 1));
				guidedResidualWriter.println();
				for (int j=0; j<geneCount; j++)
				{
					guidedResidualWriter.print("Regulator " + (j + 1));
					for (int i=0; i<geneCount; i++)
					{
						if (j == i || predicted[j][i] != 0)
							guidedResidualWriter.print("\t" + Double.NaN);
						else
						{
							int iCount = 0;
							int eCount = 0;
							for (int k=0; k<geneCount; k++)
							{
								if (predicted[k][i] == 1)
									eCount ++;
								else if (predicted[k][i] == -1)
									iCount ++;
							}

							if (modelNames[m].equals("linear"))
							{
								int[][] preds = new int[1][];
								preds[0] = new int[eCount + iCount + 1];
								double[] coefs = new double[eCount + iCount + 3];
								double[] mins = new double[eCount + iCount + 3];
								double[] maxs = new double[eCount + iCount + 3];

								for (int k=0; k<3; k++)
								{
									mins[k] = -1;
									maxs[k] = 1;
								}
								preds[0][0] = j;

								int ptr = 0;
								for (int k=0; k<geneCount; k++)
								{
									if (predicted[k][i] != 0)
									{
										preds[0][ptr+1] = k;
										if (predicted[k][i] < 0)
										{
											mins[ptr+3] = -1;
											maxs[ptr+3] = 0;
										}
										else
										{
											mins[ptr+3] = 0;
											maxs[ptr+3] = 1;
										}
										ptr ++;
									}
								}

								double guidedResidual = trajProfile.getResidual(i, preds, LinearRateModel.class, coefs, SEARCH_ITER, mins, maxs);
								guidedResiduals[m][i][j] = (coefs[2] < 0) ?-guidedResidual :guidedResidual;
								guidedResidualWriter.print("\t" + guidedResiduals[m][i][j]);
							}
							else if (modelNames[m].equals("sigmoidal"))
							{
								int[][] preds = new int[1][];
								preds[0] = new int[eCount + iCount + 1];
								double[] coefs = new double[eCount + iCount + 4];
								double[] mins = new double[eCount + iCount + 4];
								double[] maxs = new double[eCount + iCount + 4];

								for (int k=0; k<4; k++)
								{
									mins[k] = -1;
									maxs[k] = 1;
								}
								preds[0][0] = j;

								int ptr = 0;
								for (int k=0; k<geneCount; k++)
								{
									if (predicted[k][i] != 0)
									{
										preds[0][ptr+1] = k;
										if (predicted[k][i] < 0)
										{
											mins[ptr+4] = -1;
											maxs[ptr+4] = 0;
										}
										else
										{
											mins[ptr+4] = 0;
											maxs[ptr+4] = 1;
										}
										ptr ++;
									}
								}

								double guidedResidual = trajProfile.getResidual(i, preds, SigmoidalRateModel.class, coefs, SEARCH_ITER, mins, maxs);
								guidedResiduals[m][i][j] = (coefs[3] > 0) ?-guidedResidual :guidedResidual;
								guidedResidualWriter.print("\t" + guidedResiduals[m][i][j]);
							}
							else if (modelNames[m].equals("multiplicative"))
							{
								// First as inhibitor
								int[][] predsI = new int[2][];
								predsI[0] = new int[iCount + 1];
								predsI[1] = new int[eCount];
								double[] coefsI = new double[(eCount + iCount) * 2 + 4];
								double[] minsI = new double[(eCount + iCount) * 2 + 4];
								double[] maxsI = new double[(eCount + iCount) * 2 + 4];

								Arrays.fill(minsI, 0);
								Arrays.fill(maxsI, 1);
								predsI[0][0] = j;

								int ptrI0 = 1;
								int ptrI1 = 0;
								for (int k=0; k<geneCount; k++)
								{
									if (predicted[k][i] < 0)
										predsI[0][ptrI0++] = k;
									else if (predicted[k][i] > 0)
										predsI[1][ptrI1++] = k;
								}

								double guidedResidualI = trajProfile.getResidual(i, predsI, MultiplicativeRateModel.class, coefsI, SEARCH_ITER, minsI, maxsI);

								// Then as enhancer
								int[][] predsE = new int[2][];
								predsE[0] = new int[iCount];
								predsE[1] = new int[eCount + 1];
								double[] coefsE = new double[(eCount + iCount) * 2 + 4];
								double[] minsE = new double[(eCount + iCount) * 2 + 4];
								double[] maxsE = new double[(eCount + iCount) * 2 + 4];

								Arrays.fill(minsE, 0);
								Arrays.fill(maxsE, 1);
								predsI[1][0] = j;

								int ptrE0 = 0;
								int ptrE1 = 1;
								for (int k=0; k<geneCount; k++)
								{
									if (predicted[k][i] < 0)
										predsE[0][ptrE0++] = k;
									else if (predicted[k][i] > 0)
										predsE[1][ptrE1++] = k;
								}

								double guidedResidualE = trajProfile.getResidual(i, predsE, MultiplicativeRateModel.class, coefsE, SEARCH_ITER, minsE, maxsE);

								guidedResiduals[m][i][j] = (guidedResidualI < guidedResidualE) ?-guidedResidualI :guidedResidualE;
								guidedResidualWriter.print("\t" + guidedResiduals[m][i][j]);
							}
						}
					}
					guidedResidualWriter.println();
				}
				guidedResidualWriter.close();
}
			}

			// Third batch of predictions: residual < 1.5 SD from
			// mean among all potential regulators of a gene, as
			// predicted by all guided models
			predictIntersection(guidedResiduals, predictions, predicted, 3);

			// Fourth batch of predictions: residual < 1.5 SD from
			// mean among all potential regulators of a gene, as
			// predicted by some models
			predictUnion(residuals, predictions, predicted, 4);

			// Fifth batch of predictions: residual < 1.5 SD from
			// mean among all potential regulators of a gene, as
			// predicted by some guided models
			predictUnion(guidedResiduals, predictions, predicted, 5);

			// Sixth batch of predictions: prob > 0.95 in both null
			// mutants and heterozygous, and have the same direction
			{
				List<int[]> tempPredictions = new ArrayList<int[]>();
				for (int i=0; i<geneCount; i++)
					for (int j=0; j<geneCount; j++)
						if (nullReacProbs[i][j] > 0.95 && heteReacProbs[i][j] > 0.95 &&
						    (nullProfile.profile[i][j] - nullProfile.vMeans[i]) *
						    (heteProfile.profile[i][j] - heteProfile.vMeans[i]) > 0)
							tempPredictions.add(new int[] {j, i});
				double[] probs = new double[tempPredictions.size()];
				int[][] pairs = new int[tempPredictions.size()][];
				for (int i=0; i<tempPredictions.size(); i++)
				{
					pairs[i] = tempPredictions.get(i);
					probs[i] = nullReacProbs[pairs[i][1]][pairs[i][0]];
				}
				Util.coSort(probs, pairs, false);

				int newPredictions = 0;
				for (int i=0; i<pairs.length; i++)
				{
					if (predicted[pairs[i][0]][pairs[i][1]] == 0)
					{
						predictions.add("G" + (pairs[i][0] + 1) + "\tG" + (pairs[i][1] + 1));
						predicted[pairs[i][0]][pairs[i][1]] = (nullProfile.profile[pairs[i][1]][pairs[i][0]] < nullProfile.vMeans[pairs[i][1]]) ?1 :-1;
						newPredictions ++;
					}
				}
				System.out.println("Number of predictions in batch 6: " + newPredictions);
			}

			// Final batch: everything else
			{
				int remainCount = geneCount * (geneCount - 1) - predictions.size();
				double[] sds = new double[remainCount];
				int[][] pairs = new int[remainCount][];
				int ptr = 0;
				for (int i=0; i<geneCount; i++)
				{
					int count = 0;
					double sum = 0;
					double sumSq = 0;

					for (int j=0; j<geneCount; j++)
					{
						if (i == j)
							continue;

						count ++;
						sum += Math.abs(residuals[0][i][j]);
						sumSq += residuals[0][i][j] * residuals[0][i][j];
					}
					double mean = sum / count;
					double sd = Math.pow(sumSq / count - mean * mean, 0.5);

					for (int j=0; j<geneCount; j++)
					{
						if (i == j || predicted[j][i] != 0)
							continue;
						sds[ptr] = (Math.abs(residuals[0][i][j]) - mean) / sd;
						pairs[ptr] = new int[] {j, i};
						predicted[j][i] = (residuals[0][i][j] > 0) ?1 :-1;
						ptr ++;
					}
				}
				Util.coSort(sds, pairs);
				int newPredictions = 0;
				for (int i=0; i<pairs.length; i++)
				{
					predictions.add("G" + (pairs[i][0] + 1) + "\tG" + (pairs[i][1] + 1));
					predicted[pairs[i][0]][pairs[i][1]] = (nullProfile.profile[pairs[i][1]][pairs[i][0]] < nullProfile.vMeans[pairs[i][1]]) ?1 :-1;
					newPredictions ++;
				}
				System.out.println("Number of predictions in final batch: " + newPredictions);
			}

			// Output the final predictions
			File predFile = new File(resultsBase, TEAM_NAME + "_InSilico" + setID + ".txt");
			PrintWriter predWriter = new PrintWriter(new FileWriter(predFile));
			for (int i=0; i<predictions.size(); i++)
				predWriter.println(predictions.get(i) + "\t" + (1-((double)i)/predictions.size()));
			predWriter.close();
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}

	/**
	 *	Make predictions from the intersection of multiple models.
	 *	@param		r			The residuals from the models
	 *						First dimension: model
	 *						Second dimension: target gene
	 *						Third dimension: potential regulator
	 *	@param		predictions		The current list of predictions
	 *	@param		predicted		Whether a regulation has been predicted
	 *						First dimension: regulator
	 *						Second dimension: target
	 *	@param		batch			The batch number
	 */
	protected static void predictIntersection(double[][][] r, List<String> predictions, int[][] predicted, int batch)
	{
		int geneCount = r[0].length;
		Map<String, Double> tempPredictions = new HashMap<String, Double>();
		Map<String, Double> tempPredictionResiduals = new HashMap<String, Double>();
		for (int m=0; m<r.length; m++)
		{
			for (int i=0; i<geneCount; i++)
			{
				int count = 0;
				double sum = 0;
				double sumSq = 0;
				for (int j=0; j<geneCount; j++)
				{
					if (i == j)
						continue;

					double residual = r[m][i][j];
					if (Double.isNaN(residual))
						continue;

					count ++;
					sum += Math.abs(residual);
					sumSq += residual * residual;
				}
				double mean = sum / count;
				double sd = Math.pow(sumSq / count - mean * mean, 0.5);

				for (int j=0; j<geneCount; j++)
				{
					if (i == j)		// No self-interaction
						continue;

					String pair = j + "\t" + i;
					if (m != 0 && tempPredictions.get(pair) == null)
						continue;	// Not all models predict it

					double sdj = (Math.abs(r[m][i][j]) - mean) / sd;
					if (sdj < -1.5)
					{
						if (tempPredictions.get(pair) == null ||
						    tempPredictions.get(pair).doubleValue() > sdj)
						{
							tempPredictions.put(pair, new Double(sdj));
							tempPredictionResiduals.put(pair, new Double(r[m][i][j]));
						}
					}
					else			// This model does not predict it
					{
						tempPredictions.remove(pair);
						tempPredictionResiduals.remove(pair);
					}
				}
			}
		}
		String[] pairs = new String[tempPredictions.size()];
		double[] sds = new double[tempPredictions.size()];
		int ptr = 0;
		for (String pair : tempPredictions.keySet())
		{
			pairs[ptr] = pair;
			sds[ptr] = tempPredictions.get(pair).doubleValue();
			ptr ++;
		}
		Util.coSort(sds, pairs);

		int newPredictions = 0;
		for (int i=0; i<pairs.length; i++)
		{
			int regulator = Integer.parseInt(pairs[i].substring(0, pairs[i].indexOf("\t")));
			int target = Integer.parseInt(pairs[i].substring(pairs[i].indexOf("\t")+1, pairs[i].length()));
			if (predicted[regulator][target] == 0)
			{
				predictions.add("G" + (regulator + 1) + "\tG" + (target + 1));
				predicted[regulator][target] = (tempPredictionResiduals.get(pairs[i]).doubleValue() > 0) ?1 :-1;
				newPredictions ++;
			}
		}
		System.out.println("Number of predictions in batch " + batch + ": " + newPredictions);
	}

	/**
	 *	Make predictions from the union of multiple models.
	 *	@param		r			The residuals from the models
	 *						First dimension: model
	 *						Second dimension: target gene
	 *						Third dimension: potential regulator
	 *	@param		predictions		The current list of predictions
	 *	@param		predicted		Whether a regulation has been predicted
	 *						First dimension: regulator
	 *						Second dimension: target
	 *	@param		batch			The batch number
	 */
	protected static void predictUnion(double[][][] r, List<String> predictions, int[][] predicted, int batch)
	{
		int geneCount = r[0].length;
		Map<String, Double> tempPredictions = new HashMap<String, Double>();
		Map<String, Double> tempPredictionResiduals = new HashMap<String, Double>();
		for (int m=0; m<r.length; m++)
		{
			for (int i=0; i<geneCount; i++)
			{
				int count = 0;
				double sum = 0;
				double sumSq = 0;
				for (int j=0; j<geneCount; j++)
				{
					if (i == j)
						continue;

					double residual = r[m][i][j];
					if (Double.isNaN(residual))
						continue;

					count ++;
					sum += Math.abs(residual);
					sumSq += residual * residual;
				}
				double mean = sum / count;
				double sd = Math.pow(sumSq / count - mean * mean, 0.5);

				for (int j=0; j<geneCount; j++)
				{
					if (i == j)
						continue;

					String pair = j + "\t" + i;
					double sdj = (Math.abs(r[m][i][j]) - mean) / sd;
					if (sdj < -1.5)
					{
						if (tempPredictions.get(pair) == null ||
						    tempPredictions.get(pair).doubleValue() > sdj)
						{
							tempPredictions.put(pair, new Double(sdj));
							tempPredictionResiduals.put(pair, new Double(r[m][i][j]));
						}
					}
				}
			}
		}
		String[] pairs = new String[tempPredictions.size()];
		double[] sds = new double[tempPredictions.size()];
		int ptr = 0;
		for (String pair : tempPredictions.keySet())
		{
			pairs[ptr] = pair;
			sds[ptr] = tempPredictions.get(pair).doubleValue();
			ptr ++;
		}
		Util.coSort(sds, pairs);

		int newPredictions = 0;
		for (int i=0; i<pairs.length; i++)
		{
			int regulator = Integer.parseInt(pairs[i].substring(0, pairs[i].indexOf("\t")));
			int target = Integer.parseInt(pairs[i].substring(pairs[i].indexOf("\t")+1, pairs[i].length()));
			if (predicted[regulator][target] == 0)
			{
				predictions.add("G" + (regulator + 1) + "\tG" + (target + 1));
				predicted[regulator][target] = (tempPredictionResiduals.get(pairs[i]).doubleValue() > 0) ?1 :-1;
				newPredictions ++;
			}
		}
		System.out.println("Number of predictions in batch " + batch + ": " + newPredictions);
	}
}