package org.gersteinlab.regulation.model;

/**
 *	<pre>
 *	Providing methods for a rate model based on linear differential
 *	equations.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (September 1, 2008)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class LinearRateModel extends RateModel
{
	/*----------------------------------------------------------------------
	 Instance variables
	----------------------------------------------------------------------*/

	// Model data
	protected double[][] yk = null;		// Values of the potential regulators
						// First dimension: potential regulator
						// Second dimension: time point

	protected double[][] dykdt = null;	// The differentials of the potential regulators
						// First dimension: potential regulator
						// Second dimension: time point

	// Function values, gradients and Hessians of the current model at the
	// specified time points
	protected double[] yj = null;		// The estimated values of yj at the specified time points
	protected double[][] gyj = null;	// The estimated gradients of yj at the specified time points
	protected double[][][] hyj = null;	// The estimated Hessians of yj at the specified time points

	// Frequently used values
	protected double ha1d2p1 = Double.NaN;		// h * a[1] / 2 + 1
	protected double ha1_ha1d2p1_p1 = Double.NaN;	// h * a[1] * (h * a[1] / 2 + 1) + 1
	protected double h_ha1p1 = Double.NaN;		// h * (h * a[1] + 1)


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	/**
	 *	@see		org.gersteinlab.regulation.model.RateModel#RateModel(double[], double, double[][][], double[][][])
	 */
	public LinearRateModel(double[] tIn, double yj0In, double[][][] yksIn, double[][][] dykdtsIn)
	{
		super(tIn, yj0In, yksIn, dykdtsIn);
		yk = yks[0];
		dykdt = dykdts[0];
	}

	/**
	 *	Set the parameter values.
	 *	. a[0]: constant term
	 *	. a[1]: coefficient of the target gene j
	 *	. a[2..]: coefficients of the potential regulators
	 *	  (corresponding to yk[0..])
	 *	If the values in the array are changed, call setParam() again to
	 *	clear any cached intermediate results before calling the other methods.
	 *	@param		aIn			The parameter values
	 */
	public void setParam(double[] aIn)
	{
		super.setParam(aIn);
		yj = null;
		gyj = null;
		hyj = null;
		ha1d2p1 = h * a[1] / 2 + 1;
		ha1_ha1d2p1_p1 = h * a[1] * ha1d2p1 + 1;
		h_ha1p1 = h * (h * a[1] + 1);
	}

	/**
	 *	Note: to avoid array copying, the internal cached result is
	 *	returned. Calling methods should not modify the contents, or
	 *	should call setParam() first before calling this method again.
	 *	@see		org.gersteinlab.regulation.model.RateModel#f()
	 */
	public double[] f()
	{
		if (yj == null)
		{
			yj = new double[n];
			for (int i=2; i<t.length; i+=2)	// For every next time point
			{
				int index = i / 2 - 1;
				double yn = (index == 0) ?yj0 :yj[index-1];

				yj[index] =   yn
				            + h * (a[0] + a[1] * yn) * ha1d2p1;
				for (int k=0; k<yk.length; k++)
					yj[index] += h * a[k+2] * (yk[k][i-1] + h * a[1] * yk[k][i-2] / 2);
			}
		}

		return yj;
	}

	/**
	 *	Note: to avoid array copying, the internal cached result is
	 *	returned. Calling methods should not modify the contents, or
	 *	should call setParam() first before calling this method again.
	 *	@see		org.gersteinlab.regulation.model.RateModel#gf()
	 */
	public double[][] gf()
	{
		if (gyj == null)
		{
			if (yj == null)
				f();

			gyj = new double[n][];
			for (int i=2; i<t.length; i+=2)
			{
				int index = i / 2 - 1;
				gyj[index] = new double[a.length];
				double yn = (index == 0) ?yj0 :yj[index-1];

				gyj[index][0] =   h * ha1d2p1
				                + ha1_ha1d2p1_p1 * (index==0 ?0 :gyj[index-1][0]);

				gyj[index][1] =   h * h * a[0] / 2
				                + h_ha1p1 * yn;
				for (int k=0; k<yk.length; k++)
					gyj[index][1] += h * h / 2 * a[k+2] * yk[k][i-2];
				gyj[index][1] +=  ha1_ha1d2p1_p1 * (index==0 ?0 :gyj[index-1][1]);

				for (int k=0; k<yk.length; k++)
					gyj[index][k+2] =   h * (yk[k][i-1] + h * a[1] * yk[k][i-2] / 2)
					                  + ha1_ha1d2p1_p1 * (index==0 ?0 :gyj[index-1][k+2]);
			}
		}
		return gyj;
	}

	/**
	 *	Note: to avoid array copying, the internal cached result is
	 *	returned. Calling methods should not modify the contents, or
	 *	should call setParam() first before calling this method again.
	 *	@see		org.gersteinlab.regulation.model.RateModel#hf()
	 */
	public double[][][] hf()
	{
		if (hyj == null)
		{
			if (yj == null)
				f();

			if (gyj == null)
				gf();

			hyj = new double[n][][];
			for (int i=2; i<t.length; i+=2)
			{
				int index = i / 2 - 1;
				hyj[index] = new double[a.length][];
				for (int k=0; k<a.length; k++)
					hyj[index][k] = new double[a.length];
				double yn = (index == 0) ?yj0 :yj[index-1];

				for (int k1=0; k1<a.length; k1++)
				{
					for (int k2=0; k2<a.length; k2++)
					{
						if (k1 == 0 || k2 == 0)
						{
							if (k1 == 0 && k2 == 0)
								hyj[index][k1][k2] =   0;
							else if (k1 == 1 || k2 == 1)
								hyj[index][k1][k2] =   h * h / 2
								                     + h_ha1p1 * (index==0 ?0 :gyj[index-1][0])
								                     + ha1_ha1d2p1_p1 * (index==0 ?0 :hyj[index-1][0][1]);
							else
								hyj[index][k1][k2] =   ha1_ha1d2p1_p1 * (index==0 ?0 :hyj[index-1][0][k1==0 ?k2 :k1]);
						}
						else if (k1 == 1 || k2 == 1)
						{
							if (k1 == 1 && k2 == 1)
								hyj[index][k1][k2] =   h * h * yn
								                     + 2 * h_ha1p1 * (index==0 ?0 :gyj[index-1][1])
								                     + ha1_ha1d2p1_p1 * (index==0 ?0 :hyj[index-1][1][1]);
							else
								hyj[index][k1][k2] =   h * h * yn / 2
								                     + h_ha1p1 * (index==0 ?0 :gyj[index-1][k1==1 ?k2 :k1])
								                     + ha1_ha1d2p1_p1 * (index==0 ?0 :hyj[index-1][1][k1==1 ?k2 :k1]);
						}
						else
							hyj[index][k1][k2] = ha1_ha1d2p1_p1 * (index==0 ?0 :hyj[index-1][k1][k2]);
					}
				}
			}
		}
		return hyj;
	}
}