package org.gersteinlab.regulation.model;

/**
 *	<pre>
 *	Providing methods for a rate model based on multiplicative differential
 *	equations.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (September 1, 2008)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class MultiplicativeRateModel extends RateModel
{
	/*----------------------------------------------------------------------
	 Instance variables
	----------------------------------------------------------------------*/

	// Model data
	protected double[][] yk = null;		// Values of the potential inhibitors
						// First dimension: potential inhibitor
						// Second dimension: time point

	protected double[][] dykdt = null;	// The differentials of the potential inhibitors
						// First dimension: potential inhibitor
						// Second dimension: time point

	protected double[][] yl = null;		// Values of the potential enhancers
						// First dimension: potential inhibitor
						// Second dimension: time point

	protected double[][] dyldt = null;	// The differentials of the potential enhancers
						// First dimension: potential inhibitor
						// Second dimension: time point

	protected int kCount = -1;		// Number of inhibitors in the current model
	protected int lCount = -1;		// Number of enhancers in the current model

	// Model parameters
	double aj0 = Double.NaN;		// Basal expression rate
	double aj1 = Double.NaN;		// Degradation rate
	double[] bjk = null;			// Rate constant of inhibitors k
	double[] cjk = null;			// Sigmoidal constant of inhibitors k
	double[] bjl = null;			// Rate constant of enhancers l
	double[] cjl = null;			// Sigmoidal constant of enhancers l

	// Function values, gradients and Hessians of the current model at the
	// specified time points
	protected double[] yj = null;		// The estimated values of yj at the specified time points
	protected double[][] gyj = null;	// The estimated gradients of yj at the specified time points
	protected double[][][] hyj = null;	// The estimated Hessians of yj at the specified time points

	// Frequently used values
	protected double[] bigProduct = null;


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	/**
	 *	@see		org.gersteinlab.regulation.model.RateModel#RateModel(double[], double, double[][][], double[][][])
	 */
	public MultiplicativeRateModel(double[] tIn, double yj0In, double[][][] yksIn, double[][][] dykdtsIn)
	{
		super(tIn, yj0In, yksIn, dykdtsIn);
		yk = yks[0];
		dykdt = dykdts[0];
		yl = yks[1];
		dyldt = dykdts[1];
		kCount = yk.length;
		lCount = yl.length;
	}

	/**
	 *	Set the parameter values.
	 *	. a[0]: a_{j0}, basal expression rate
	 *	. a[1]: a_{j1}, degradation rate
	 *	. a[2,4,...,2x]: b_{jk}, rate constant of inhibitor k
	 *	. a[3,5,...,2x+1]: c_{jk}, sigmoidal constant of inhibitor k
	 *	. a[2x+2,2x+4,...]: b_{jl}, rate constant of enhancer l
	 *	. a[2x+3,2x+5,...]: c_{jl}, sigmoidal constant of enhancer l
	 *
	 *	If the values in the array are changed, call setParam() again to
	 *	clear any cached intermediate results before calling the other
	 *	methods.
	 *
	 *	@param		aIn			The parameter values
	 */
	public void setParam(double[] aIn)
	{
		super.setParam(aIn);

		aj0 = aIn[0];
		aj1 = aIn[1];
		bjk = new double[kCount];
		cjk = new double[kCount];
		for (int k=0; k<kCount; k++)
		{
			bjk[k] = aIn[2 + 2 * k];
			cjk[k] = aIn[2 + 2 * k + 1];
		}
		bjl = new double[lCount];
		cjl = new double[lCount];
		for (int l=0; l<lCount; l++)
		{
			bjl[l] = aIn[2 * 2 * kCount + 2 * l];
			cjl[l] = aIn[2 * 2 * kCount + 2 * l + 1];
		}

		yj = null;
		gyj = null;
		hyj = null;
	}

	/**
	 *	Note: to avoid array copying, the internal cached result is
	 *	returned. Calling methods should not modify the contents, or
	 *	should call setParam() first before calling this method again.
	 *	@see		org.gersteinlab.regulation.model.RateModel#f()
	 */
	public double[] f()
	{
		if (yj == null)
		{
			bigProduct = new double[t.length];
			for (int i=0; i<t.length; i++)
			{
				bigProduct[i] = 1;
				for (int k=0; k<kCount; k++)
					bigProduct[i] *= bjk[k] / (Math.pow(yk[k][i], cjk[k]) + bjk[k]);
				for (int l=0; l<lCount; l++)
					bigProduct[i] *= Math.pow(yl[l][i], cjl[l]) / (Math.pow(yl[l][i], cjl[l]) + bjl[l]);
			}

			yj = new double[n];
			for (int i=2; i<t.length; i+=2)	// For every next time point
			{
				int index = i / 2 - 1;
				double yn = (index == 0) ?yj0 :yj[index-1];
				if (yn < 0)
					yn = 0;

				yj[index] =   h * h * aj1 * aj1 / 2 * yn * yn
				            + (1 - h * aj1) * yn
				            + h * aj0 * bigProduct[i-1]
				            - h * h * aj0 * aj1 / 2 * bigProduct[i-2];
			}
		}

		return yj;
	}

	/**
	 *	Note: to avoid array copying, the internal cached result is
	 *	returned. Calling methods should not modify the contents, or
	 *	should call setParam() first before calling this method again.
	 *	@see		org.gersteinlab.regulation.model.RateModel#gf()
	 */
	public double[][] gf()
	{
		if (gyj == null)
		{
			if (yj == null)
				f();

			gyj = new double[n][];
			for (int i=2; i<t.length; i+=2)
			{
				int index = i / 2 - 1;
				gyj[index] = new double[a.length];
				double yn = (index == 0) ?yj0 :yj[index-1];
				if (yn < 0)
					yn = 0;

				gyj[index][0] =   (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :gyj[index-1][0])
				                + h * bigProduct[i-1]
				                - h * h * aj1 / 2 * bigProduct[i-2];

				gyj[index][1] =   h * h * aj1 * yn * yn
				                - h * yn
				                + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :gyj[index-1][1])
				                - h * h * aj0 / 2 * bigProduct[i-2];

				for (int k=0; k<kCount; k++)
				{
					int base = 2 + k * 2;
					gyj[index][base]   =   (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :gyj[index-1][base])
					                     + h * aj0               * Math.pow(yk[k][i-1], cjk[k]) / bjk[k] / (Math.pow(yk[k][i-1], cjk[k]) + bjk[k]) * bigProduct[i-1]
					                     - h * h * aj0 * aj1 / 2 * Math.pow(yk[k][i-2], cjk[k]) / bjk[k] / (Math.pow(yk[k][i-2], cjk[k]) + bjk[k]) * bigProduct[i-2];

					gyj[index][base+1] =   (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :gyj[index-1][base+1])
					                     - h * aj0               * Math.pow(yk[k][i-1], cjk[k]) * Math.log(yk[k][i-1]) / (Math.pow(yk[k][i-1], cjk[k]) + bjk[k]) * bigProduct[i-1]
					                     + h * h * aj0 * aj1 / 2 * Math.pow(yk[k][i-2], cjk[k]) * Math.log(yk[k][i-2]) / (Math.pow(yk[k][i-2], cjk[k]) + bjk[k]) * bigProduct[i-2];
				}

				for (int l=0; l<lCount; l++)
				{
					int base = 2 + kCount * 2 + l * 2;
					gyj[index][base]   =   (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :gyj[index-1][base])
					                     - h * aj0               / (Math.pow(yl[l][i-1], cjl[l]) + bjl[l]) * bigProduct[i-1]
					                     + h * h * aj0 * aj1 / 2 / (Math.pow(yl[l][i-2], cjl[l]) + bjl[l]) * bigProduct[i-2];

					gyj[index][base+1] =   (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :gyj[index-1][base+1])
					                     + h * aj0 * bjl[l]               * Math.log(yl[l][i-1]) / (Math.pow(yl[l][i-1], cjl[l]) + bjl[l]) * bigProduct[i-1]
					                     - h * h * aj0 * aj1 / 2 * bjl[l] * Math.log(yl[l][i-2]) / (Math.pow(yl[l][i-2], bjl[l]) + bjl[l]) * bigProduct[i-2];
				}
			}
		}
		return gyj;
	}

	/**
	 *	Note: to avoid array copying, the internal cached result is
	 *	returned. Calling methods should not modify the contents, or
	 *	should call setParam() first before calling this method again.
	 *	@see		org.gersteinlab.regulation.model.RateModel#hf()
	 */
	public double[][][] hf()
	{
		if (hyj == null)
		{
			if (yj == null)
				f();

			if (gyj == null)
				gf();

			hyj = new double[n][][];
			for (int i=2; i<t.length; i+=2)
			{
				int index = i / 2 - 1;
				hyj[index] = new double[a.length][];
				for (int k=0; k<a.length; k++)
					hyj[index][k] = new double[a.length];
				double yn = (index == 0) ?yj0 :yj[index-1];
				if (yn < 0)
					yn = 0;

				for (int k1=0; k1<a.length; k1++)
				{
					for (int k2=0; k2<a.length; k2++)
					{
						if (k1 == 0 || k2 == 0)		// aj0
						{
							int kk = (k1==0) ?k2 :k1;

							if (kk == 0)			// aj0
								hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :Math.pow(gyj[index-1][0], 2))
								                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][0][0]);

							else if (kk == 1)		// aj1
								hyj[index][k1][k2] =   (2 * h * h * aj1 * yn + h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][1]) - h) * (index==0 ?0 :gyj[index-1][1])
								                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 : hyj[index-1][0][1])
								                     - h * h / 2 * bigProduct[i-2];

							else if (kk < 2 + kCount * 2)
							{
								int k = (kk - 2) / 2;
								if (kk % 2 == 0)	// bjk
									hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][0]*gyj[index-1][kk])
									                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][0][kk])
									                     + h               * Math.pow(yk[k][i-1], cjk[k]) / bjk[k] / (Math.pow(yk[k][i-1], cjk[k]) + bjk[k]) * bigProduct[i-1]
									                     - h * h * aj1 / 2 * Math.pow(yk[k][i-2], cjk[k]) / bjk[k] / (Math.pow(yk[k][i-2], cjk[k]) + bjk[k]) * bigProduct[i-2];

								else			// cjk
									hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][0]*gyj[index-1][kk])
									                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][0][kk])
									                     - h               * Math.pow(yk[k][i-1], cjk[k]) * Math.log(yk[k][i-1]) / (Math.pow(yk[k][i-1], cjk[k]) + bjk[k]) * bigProduct[i-1]
									                     + h * h * aj1 / 2 * Math.pow(yk[k][i-2], cjk[k]) * Math.log(yk[k][i-2]) / (Math.pow(yk[k][i-2], cjk[k]) + bjk[k]) * bigProduct[i-2];
							}
							else
							{
								int l = (kk - 2 - kCount * 2) / 2;
								if (kk % 2 == 0)	// bjl
									hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][0]*gyj[index-1][kk])
									                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][0][kk])
									                     - h               / (Math.pow(yl[l][i-1], cjl[l]) + bjl[l]) * bigProduct[i-1]
									                     + h * h * aj1 / 2 / (Math.pow(yl[l][i-2], cjl[l]) + bjl[l]) * bigProduct[i-2];

								else			// cjl
									hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][0]*gyj[index-1][kk])
									                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][0][kk])
									                     + h               * bjl[l] * Math.log(yl[l][i-1]) / (Math.pow(yl[l][i-1], cjl[l]) + bjl[l]) * bigProduct[i-1]
									                     - h * h * aj1 / 2 * bjl[l] * Math.log(yl[l][i-2]) / (Math.pow(yl[l][i-2], cjl[l]) + bjl[l]) * bigProduct[i-2];
							}
						}

						else if (k1 == 1 || k2 == 1)	// aj1
						{
							int kk = (k1==1) ?k2 :k1;

							if (kk == 1)			// aj1
								hyj[index][k1][k2] =   h * h * yn * yn
								                     + 4 * h * h * aj1 * yn * (index==0 ?0 :gyj[index-1][1])
								                     - 2 * h * (index==0 ?0 :gyj[index-1][1])
								                     + h * h * aj1 * aj1 * (index==0 ?0 :Math.pow(gyj[index-1][1], 2))
								                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][1][1]);

							else if (kk < 2 + kCount * 2)
							{
								int k = (kk - 2) / 2;

								if (kk % 2 == 0)	// bjk
									hyj[index][k1][k2] =   2 * h * h * aj1 * yn * (index==0 ?0 :gyj[index-1][kk])
									                     - h * (index==0 ?0 :gyj[index-1][kk])
									                     + h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][1]*gyj[index-1][kk])
									                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][1][kk])
									                     - h * h * aj0 / 2 * Math.pow(yk[k][i-2], cjk[k]) / bjk[k] / (Math.pow(yk[k][i-2], cjk[k]) + bjk[k]) * bigProduct[i-2];

								else			// cjk
									hyj[index][k1][k2] =   2 * h * h * aj1 * yn * (index==0 ?0 :gyj[index-1][kk])
									                     - h * (index==0 ?0 :gyj[index-1][kk])
									                     + h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][1]*gyj[index-1][kk])
									                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][1][kk])
									                     + h * h * aj0 / 2 * Math.pow(yk[k][i-2], cjk[k]) * Math.log(yk[k][i-2]) / (Math.pow(yk[k][i-2], cjk[k]) + bjk[k]) * bigProduct[i-2];
							}

							else
							{
								int l = (kk - 2 - kCount * 2) / 2;

								if (kk % 2 == 0)	// bjl
									hyj[index][k1][k2] =   2 * h * h * aj1 * yn * (index==0 ?0 :gyj[index-1][kk])
									                     - h * (index==0 ?0 :gyj[index-1][kk])
									                     + h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][1]*gyj[index-1][kk])
									                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][1][kk])
									                     + h * h * aj0 / 2 / (Math.pow(yl[l][i-2], cjl[l]) + bjl[l]) * bigProduct[i-2];

								else			// cjl
									hyj[index][k1][k2] =   2 * h * h * aj1 * yn * (index==0 ?0 :gyj[index-1][kk])
									                     - h * (index==0 ?0 :gyj[index-1][kk])
									                     + h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][1]*gyj[index-1][kk])
									                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][1][kk])
									                     - h * h * aj0 / 2 * bjl[l] * Math.log(yl[l][i-2]) / (Math.pow(yl[l][i-2], cjl[l]) + bjl[l]) * bigProduct[i-2];
							}
						}

						else if ((k1 < 2 + kCount * 2 && k1 % 2 == 0) ||
						         (k2 < 2 + kCount * 2 && k2 % 2 == 0))
						         			// bjk
						{
							int kk = (k1 < 2 + kCount * 2 && k1 % 2 == 0) ?k1 :k2;
							int k = (kk - 2) / 2;
							int kkp = (k1 < 2 + kCount * 2 && k1 % 2 == 0) ?k2 :k1;

							if (k1 == k2)			// bjk
								hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :Math.pow(gyj[index-1][kk], 2))
								                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][kk][kk])
								                     - 2 * h * aj0       * Math.pow(yk[k][i-1], cjk[k]) / bjk[k] / Math.pow(Math.pow(yk[k][i-1], cjk[k]) + bjk[k], 2) * bigProduct[i-1]
								                     + h * h * aj0 * aj1 * Math.pow(yk[k][i-2], cjk[k]) / bjk[k] / Math.pow(Math.pow(yk[k][i-2], cjk[k]) + bjk[k], 2) * bigProduct[i-2];

							else if (kkp < 2 + kCount * 2)
							{
								int kp = (kkp - 2) / 2;

								if (kkp % 2 == 0)	// bjkp
									hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][kk]*gyj[index-1][kkp])
									                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][kk][kkp])
									                     + h * aj0               * Math.pow(yk[k ][i-1], cjk[k ]) / bjk[k ] / (Math.pow(yk[k ][i-1], cjk[k ]) + bjk[k ])
									                                             * Math.pow(yk[kp][i-1], cjk[kp]) / bjk[kp] / (Math.pow(yk[kp][i-1], cjk[kp]) + bjk[kp]) * bigProduct[i-1]
									                     - h * h * aj0 * aj1 / 2 * Math.pow(yk[k ][i-2], cjk[k ]) / bjk[k ] / (Math.pow(yk[k ][i-2], cjk[k ]) + bjk[k ])
									                                             * Math.pow(yk[kp][i-2], cjk[kp]) / bjk[kp] / (Math.pow(yk[kp][i-2], cjk[kp]) + bjk[kp]) * bigProduct[i-2];
								else			// cjkp
									hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][kk]*gyj[index-1][kkp])
									                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][kk][kkp])
									                     - h * aj0               * Math.pow(yk[k][i-1], cjk[k]) / bjk[k] / (Math.pow(yk[k][i-1], cjk[k]) + bjk[k]) * Math.pow(yk[kp][i-1], cjk[kp]) * Math.log(yk[kp][i-1]) / (Math.pow(yk[kp][i-1], cjk[kp]) + bjk[kp]) * bigProduct[i-1]
									                     + h * h * aj0 * aj1 / 2 * Math.pow(yk[k][i-2], cjk[k]) / bjk[k] / (Math.pow(yk[k][i-2], cjk[k]) + bjk[k]) * Math.pow(yk[kp][i-2], cjk[kp]) * Math.log(yk[kp][i-2]) / (Math.pow(yk[kp][i-2], cjk[kp]) + bjk[kp]) * bigProduct[i-2];
							}

							else
							{
								int ll = kkp;
								int l = (ll - 2 - kCount * 2) / 2;

								if (ll % 2 == 0)	// bjl
									hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][kk]*gyj[index-1][ll])
									                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][kk][ll])
									                     - h * aj0               * Math.pow(yk[k][i-1], cjk[k]) / bjk[k] / (Math.pow(yk[k][i-1], cjk[k]) + bjk[k]) / (Math.pow(yl[l][i-1], cjl[l]) + bjl[l]) * bigProduct[i-1]
									                     + h * h * aj0 * aj1 / 2 * Math.pow(yk[k][i-2], cjk[k]) / bjk[k] / (Math.pow(yk[k][i-2], cjk[k]) + bjk[k]) / (Math.pow(yl[l][i-2], cjl[l]) + bjl[l]) * bigProduct[i-2];

								else			// cjl
									hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][kk]*gyj[index-1][ll])
									                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][kk][ll])
									                     + h * aj0               * Math.pow(yk[k][i-1], cjk[k]) / bjk[k] / (Math.pow(yk[k][i-1], cjk[k]) + bjk[k]) * bjl[l] * Math.log(yl[l][i-1]) / (Math.pow(yl[l][i-1], cjl[l]) + bjl[l]) * bigProduct[i-1]
									                     - h * h * aj0 * aj1 / 2 * Math.pow(yk[k][i-2], cjk[k]) / bjk[k] / (Math.pow(yk[k][i-2], cjk[k]) + bjk[k]) * bjl[l] * Math.log(yl[l][i-2]) / (Math.pow(yl[l][i-2], cjl[l]) + bjl[l]) * bigProduct[i-2];
							}
						}

						else if (k1 < 2 + kCount * 2 ||
						         k2 < 2 + kCount * 2)	// cjk
						{
							int kk = (k1 < 2 + kCount * 2 && k1 % 2 == 0) ?k1 :k2;
							int k = (kk - 2) / 2;
							int kkp = (k1 < 2 + kCount * 2 && k1 % 2 == 0) ?k2 :k1;

							if (k1 == k2)			// cjk
								hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :Math.pow(gyj[index-1][kk], 2))
								                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][kk][kk])
								                     - h * aj0               * Math.pow(yk[k][i-1], cjk[k]) * Math.pow(Math.log(yk[k][i-1]), 2) * (bjk[k] - Math.pow(yk[k][i-1], cjk[k])) / Math.pow(Math.pow(yk[k][i-1], cjk[k]) + bjk[k], 2) * bigProduct[i-1]
								                     + h * h * aj0 * aj1 / 2 * Math.pow(yk[k][i-2], cjk[k]) * Math.pow(Math.log(yk[k][i-2]), 2) * (bjk[k] - Math.pow(yk[k][i-2], cjk[k])) / Math.pow(Math.pow(yk[k][i-2], cjk[k]) + bjk[k], 2) * bigProduct[i-2];

							else if (kkp < 2 + kCount * 2)	// cjkp
							{
								int kp = (kkp - 2) / 2;
								hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][kk]*gyj[index-1][kkp])
								                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][kk][kkp])
								                     + h * aj0               * Math.pow(yk[k ][i-1], cjk[k ]) * Math.log(yk[k ][i-1]) / (Math.pow(yk[k ][i-1], cjk[k ]) + bjk[k ])
								                                             * Math.pow(yk[kp][i-1], cjk[kp]) * Math.log(yk[kp][i-1]) / (Math.pow(yk[kp][i-1], cjk[kp]) + bjk[kp]) * bigProduct[i-1]
								                     - h * h * aj0 * aj1 / 2 * Math.pow(yk[k ][i-2], cjk[k ]) * Math.log(yk[k ][i-2]) / (Math.pow(yk[k ][i-2], cjk[k ]) + bjk[k ])
								                                             * Math.pow(yk[kp][i-2], cjk[kp]) * Math.log(yk[kp][i-2]) / (Math.pow(yk[kp][i-2], cjk[kp]) + bjk[kp]) * bigProduct[i-2];
							}

							else
							{
								int ll = kkp;
								int l = (ll - 2 - kCount * 2) / 2;

								if (ll % 2 == 0)	// bjl
									hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][kk]*gyj[index-1][ll])
									                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][kk][ll])
									                     + h * aj0               * Math.pow(yk[k][i-1], cjk[k]) * Math.log(yk[k][i-1]) / (Math.pow(yk[k][i-1], cjk[k]) + bjk[k]) / (Math.pow(yl[l][i-1], cjl[l]) + bjl[l]) * bigProduct[i-1]
									                     - h * h * aj0 * aj1 / 2 * Math.pow(yk[k][i-2], cjk[k]) * Math.log(yk[k][i-2]) / (Math.pow(yk[k][i-2], cjk[k]) + bjk[k]) / (Math.pow(yl[l][i-2], cjl[l]) + bjl[l]) * bigProduct[i-2];

								else			// cjl
									hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][kk]*gyj[index-1][ll])
									                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][kk][ll])
									                     - h * aj0               * Math.pow(yk[k][i-1], cjk[k]) * Math.log(yk[k][i-1]) / (Math.pow(yk[k][i-1], cjk[k]) + bjk[k]) * bjl[l] * Math.log(yl[l][i-1]) / (Math.pow(yl[l][i-1], cjl[l]) + bjl[l]) * bigProduct[i-1]
									                     + h * h * aj0 * aj1 / 2 * Math.pow(yk[k][i-2], cjk[k]) * Math.log(yk[k][i-2]) / (Math.pow(yk[k][i-2], cjk[k]) + bjk[k]) * bjl[l] * Math.log(yl[l][i-2]) / (Math.pow(yl[l][i-2], cjl[l]) + bjl[l]) * bigProduct[i-2];
							}
						}

						else if (k1 % 2 == 0 ||
						         k2 % 2 == 0)		// bjl
						{
							int ll = (k1 % 2 == 0) ?k1 :k2;
							int l = (ll - 2 - kCount * 2) / 2;

							if (k1 == k2)			// bjl
								hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :Math.pow(gyj[index-1][ll], 2))
								                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][ll][ll])
								                     + 2 * h * aj0       / Math.pow(Math.pow(yl[l][i-1], cjl[l]) + bjl[l], 2) * bigProduct[i-1]
								                     - h * h * aj0 * aj1 / Math.pow(Math.pow(yl[l][i-2], cjl[l]) + bjl[l], 2) * bigProduct[i-2];

							else
							{
								int llp = (k1 % 2 == 0) ?k2 :k1;
								int lp = (llp - 2 - kCount * 2) / 2;

								if (llp % 2 == 0)	// bjlp
									hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][ll]*gyj[index-1][llp])
									                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][ll][llp])
									                     + h * aj0               / (Math.pow(yl[l ][i-1], cjl[l ]) + bjl[l ])
									                                             / (Math.pow(yl[lp][i-1], cjl[lp]) + bjl[lp]) * bigProduct[i-1]
									                     - h * h * aj0 * aj1 / 2 / (Math.pow(yl[l ][i-2], cjl[l ]) + bjl[l ])
									                                             / (Math.pow(yl[lp][i-2], cjl[lp]) + bjl[lp]) * bigProduct[i-2];

								else			// cjlp
									hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][ll]*gyj[index-1][llp])
									                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][ll][llp])
									                     - h * aj0               / (Math.pow(yl[l][i-1], cjl[l]) + bjl[l]) * bjl[l] * Math.log(yl[l][i-1]) / (Math.pow(yl[l][i-1], cjl[l]) + bjl[l]) * bigProduct[i-1]
									                     + h * h * aj0 * aj1 / 2 / (Math.pow(yl[l][i-2], cjl[l]) + bjl[l]) * bjl[l] * Math.log(yl[l][i-2]) / (Math.pow(yl[l][i-2], cjl[l]) + bjl[l]) * bigProduct[i-2];
							}
						}

						else				// cjl
						{
							int ll = k1;
							int l = (ll - 2 - kCount * 2) / 2;

							if (k1 == k2)			// cjl
								hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :Math.pow(gyj[index-1][ll], 2))
								                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][ll][ll])
								                     + h * aj0               * bjl[l] * Math.pow(Math.log(yl[l][i-1]), 2) * (bjl[l] - Math.pow(yl[l][i-1], cjl[l])) / Math.pow(Math.pow(yl[l][i-1], cjl[l]) + bjl[l], 2) * bigProduct[i-1]
								                     - h * h * aj0 * aj1 / 2 * bjl[l] * Math.pow(Math.log(yl[l][i-2]), 2) * (bjl[l] - Math.pow(yl[l][i-2], cjl[l])) / Math.pow(Math.pow(yl[l][i-2], cjl[l]) + bjl[l], 2) * bigProduct[i-2];

							else				// cjlp
							{
								int llp = k2;
								int lp = (llp - 2 - kCount * 2) / 2;

								hyj[index][k1][k2] =   h * h * aj1 * aj1 * (index==0 ?0 :gyj[index-1][ll]*gyj[index-1][llp])
								                     + (h * h * aj1 * aj1 * yn - h * aj1 + 1) * (index==0 ?0 :hyj[index-1][ll][llp])
								                     + h * aj0               * bjl[l ] * Math.log(yl[l ][i-1]) / (Math.pow(yl[l ][i-1], cjl[l ]) + bjl[l ])
								                                             * bjl[lp] * Math.log(yl[lp][i-1]) / (Math.pow(yl[lp][i-1], cjl[lp]) + bjl[lp]) * bigProduct[i-1]
								                     - h * h * aj0 * aj1 / 2 * bjl[l ] * Math.log(yl[l ][i-2]) / (Math.pow(yl[l ][i-2], cjl[l ]) + bjl[l ])
								                                             * bjl[lp] * Math.log(yl[lp][i-2]) / (Math.pow(yl[lp][i-2], cjl[lp]) + bjl[lp]) * bigProduct[i-2];
							}
						}
					}
				}
			}
		}
		return hyj;
	}
}