package org.gersteinlab.regulation.model;

import org.gersteinlab.regulation.numeric.*;

/**
 *	<pre>
 *	A class that represents a polynomial.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (August 29, 2008)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class Polynomial extends DoublyDifferentiableFunction
{
	/*----------------------------------------------------------------------
	 Instance variables
	----------------------------------------------------------------------*/

	protected double[] coefs = null;	// Coefficients
	protected double[] dCoefs = null;	// Coefficients for first derivative
	protected double[] d2Coefs = null;	// Coefficients for second derivative


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	/**
	 *	Create a polynomial with a set of coefficients.
	 *	@param		coefsIn			The coefficients
	 */
	public Polynomial(double[] coefsIn)
	{
		coefs = coefsIn;
	}

	/**
	 *	@see		org.gersteinlab.regulation.numeric.Function#f(double)
	 */
	public double f(double x)
	{
		double result = 0;
		for (int i=coefs.length-1; i>=0; i--)
			result = result * x + coefs[i];
		return result;
	}

	/**
	 *	@see		org.gersteinlab.regulation.numeric.DifferentiableFunction#f(double)
	 */
	public double df(double x)
	{
		if (dCoefs == null)
		{
			int n = coefs.length - 1;
			if (n == 0)
				dCoefs = new double[] {0};
			else
			{
				dCoefs = new double[n];
				for (int i=0; i<n; i++)
					dCoefs[i] = coefs[i+1] * (i+1);
			}
		}

		double result = 0;
		for (int i=dCoefs.length-1; i>=0; i--)
			result = result * x + dCoefs[i];
		return result;
	}

	/**
	 *	@see		org.gersteinlab.regulation.numeric.DoublyDifferentiableFunction#f(double)
	 */
	public double d2f(double x)
	{
		if (d2Coefs == null)
		{
			int n = coefs.length - 1;
			if (n == 0 || n == 1)
				d2Coefs = new double[] {0};
			else
			{
				d2Coefs = new double[n-1];
				for (int i=0; i<n-1; i++)
					d2Coefs[i] = coefs[i+2] * (i+1) * (i+2);
			}
		}

		double result = 0;
		for (int i=d2Coefs.length-1; i>=0; i--)
			result = result * x + d2Coefs[i];
		return result;
	}
}