package org.gersteinlab.regulation.model;

import org.gersteinlab.regulation.numeric.*;

/**
 *	<pre>
 *	A function that equals the total squared difference between a polynomial
 *	and some target values at a set of points, treating the coefficients of
 *	the polynomial as the variables.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (August 29, 2008)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class PolynomialSqrDiff extends DoublyDifferentiableMultivariateFunction
{
	/*----------------------------------------------------------------------
	 Instance variables
	----------------------------------------------------------------------*/

	double[] t = null;			// The given points
	double[] y = null;			// The given target values


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	/**
	 *	Create the function with the given set of points and target
	 *	values.
	 *	@param		tIn			The points
	 *	@param		yIn			The target values
	 */
	public PolynomialSqrDiff(double[] tIn, double[] yIn)
	{
		t = tIn;
		y = yIn;
	}

	/**
	 *	Evaluate the total squared difference at a point (i.e., a set of
	 *	polynomial coefficients).
	 *	@param		a			The coefficients
	 *	@return					The squred difference
	 */
	public double f(double[] a)
	{
		Polynomial p = new Polynomial(a);
		int n = t.length;

		double result = 0;
		for (int i=0; i<n; i++)
			result += Math.pow(p.f(t[i]) - y[i], 2);
		return result;
	}

	/**
	 *	Evaluate the gradient of the function at a point (i.e., a set of
	 *	polynomial coefficients).
	 *	@param		a			The point
	 *	@return					The gradient of the function at the point
	 */
	public double[] gf(double[] a)
	{
		Polynomial p = new Polynomial(a);
		int n = t.length;
		int d = a.length;

		double temp[] = new double[n];
		for (int i=0; i<n; i++)
		{
			for (int j=d-1; j>=0; j--)
				temp[i] = temp[i] * t[i] + a[j];
			temp[i] -= y[i];
			temp[i] *= 2;
		}

		double[] result = new double[d];
		for (int i=0; i<n; i++)
			for (int j=0; j<d; j++)
				result[j] += temp[i] * Math.pow(t[i], j);
		return result;
	}

	/**
	 *	Evaluate the Hessian of the function at a point (i.e., a set of
	 *	polynomial coefficients).
	 *	@param		a			The point
	 *	@return					The Hessian of the function at the point
	 */
	public double[][] hf(double[] a)
	{
		Polynomial p = new Polynomial(a);
		int n = t.length;
		int d = a.length;

		double[][] result = new double[d][];
		for (int j=0; j<d; j++)
			result[j] = new double[d];

		for (int j=0; j<d; j++)
			for (int k=0; k<d; k++)
				for (int i=0; i<n; i++)
					result[j][k] += 2 * Math.pow(t[i], j + k);
		return result;
	}
}