package org.gersteinlab.regulation.model;

/**
 *	<pre>
 *	Providing methods for a rate model based on differential equations. The
 *	goal is the estimate the expression levels of yj at different time
 *	points given its initial value at time 0 and the expression levels and
 *	rates of some potential regulators.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (September 1, 2008)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public abstract class RateModel
{
	/*----------------------------------------------------------------------
	 Instance variables
	----------------------------------------------------------------------*/

	protected double[] t = null;		// Time points
						// The points t[2], t[4], ... are used to evaluate the model,
						// while the remaining points are used to estimate the slopes
						// of Runge-Kutta

	protected double yj0 = Double.NaN;	// The initial value at t[0]

	protected double[][][] yks = null;	// Values of the potential regulators
						// First dimension: regulator set
						// Second dimension: potential regulator
						// Third dimension: time point

	protected double[][][] dykdts = null;	// The differentials of the potential regulators
						// First dimension: regulator set
						// Second dimension: potential regulator
						// Third dimension: time point

	protected double h = Double.NaN;	// Step size
	protected int n = -1;			// Number of specified time points for evaluating the model

	protected double[] a = null;		// The parameter values


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	/**
	 *	Create the model with the given set of time points and target
	 *	values and differentials.
	 *	@param		tIn			The set of time points
	 *	@param		yj0In			The initial target value
	 *	@param		yksIn			The values of the potential regulators
	 *						First dimension: potential regulator
	 *						Second dimension: time point
	 *	@param		dykdtsIn		The differentials of the potential regulators
	 *						First dimension: potential regulator
	 *						Second dimension: time point
	 */
	public RateModel(double[] tIn, double yj0In, double[][][] yksIn, double[][][] dykdtsIn)
	{
		t = tIn;
		yj0 = yj0In;
		yks = yksIn;
		dykdts = dykdtsIn;
		h = t[2] - t[0];
		n = (tIn.length - 1) / 2;
		if (n < 0)
			n = 0;
	}

	/**
	 *	Set the parameter values.
	 *	@param		aIn			The parameter values
	 */
	public void setParam(double[] aIn)
	{
		a = aIn;
	}

	/**
	 *	Return the estimated values at the specified time points.
	 *	@return					The estimated values at the specified time points
	 */
	public abstract double[] f();

	/**
	 *	Return the estimated gradients at the specified time points with
	 *	respect to the parameters.
	 *	@return					The estimated gradients at the specified time points
	 *						First dimension: time point
	 *						Second dimension: parameter 
	 */
	public abstract double[][] gf();

	/**
	 *	Return the estimated Hessians at the specified time points with
	 *	respect to the parameters.
	 *	@return					The estimated Hessians at the specified time points
	 *						First dimension: time point
	 *						Second and third dimensions: parameter
	 */
	public abstract double[][][] hf();
}