package org.gersteinlab.regulation.model;

import org.gersteinlab.regulation.numeric.*;

import java.lang.reflect.*;

/**
 *	<pre>
 *	A function that equals the total squared difference between the
 *	predicted value of a differential equation rate model and some target
 *	values at a set of time points, treating the parameters of the model as
 *	the variables.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (September 1, 2008)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class RateModelSqrDiff extends DoublyDifferentiableMultivariateFunction
{
	/*----------------------------------------------------------------------
	 Instance variables
	----------------------------------------------------------------------*/

	double[][] yj = null;			// The sets of given target values
						// First dimension: experiment
						// Second dimension: time point

	RateModel[] models = null;		// The rate models for the different experiments


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	/**
	 *	Create the function with the given set of points and target
	 *	values and differentials.
	 *	@param		tIn			The sets of points
	 *	@param		yjIn			The sets of target values
	 *						First dimension: experiment
	 *						Second dimension: time point
	 *	@param		yksIn			The sets of predictors
	 *						First dimention: predictor set
	 *						Second dimension: predictor
	 *						Third dimension: experiment
	 *						Fourth dimension: time point
	 *	@param		dykdtsIn		The sets of differentials of the predictors
	 *						First dimension: predictor set
	 *						Second dimension: predictor
	 *						Third dimension: experiment
	 *						Fourth dimension: time point
	 *	@param		modelClass		The class of the rate model
	 */
	public RateModelSqrDiff(double[] tIn, double[][] yjIn, double[][][][] yksIn, double[][][][] dykdtsIn, Class<? extends RateModel> modelClass)
	throws InstantiationException,
	       IllegalAccessException,
	       IllegalArgumentException,
	       InvocationTargetException,
	       NoSuchMethodException
	{
		yj = yjIn;

		int expCount = yjIn.length;
		int predSetCount = yksIn.length;
		models = new RateModel[expCount];
		for (int e=0; e<expCount; e++)			// One model per experiment
		{
			Constructor modelConstructor = modelClass.getConstructor(double[].class, double.class, double[][][].class, double[][][].class);
			double[][][] yks = new double[predSetCount][][];
			double[][][] dykdts = new double[predSetCount][][];
			for (int k=0; k<predSetCount; k++)	// For each predictor set
			{
				int predCount = yksIn[k].length;
				yks[k] = new double[predCount][];
				dykdts[k] = new double[predCount][];
				for (int l=0; l<predCount; l++)	// For each predictor
				{
					yks[k][l] = yksIn[k][l][e];
					dykdts[k][l] = dykdtsIn[k][l][e];
				}
			}
			models[e] = (RateModel)modelConstructor.newInstance(tIn, yjIn[e][0], yks, dykdts);
		}
	}

	/**
	 *	Set the parameter values of the model.
	 *	@param		a			The parameter values
	 */
	public void set(double[] a)
	{
		super.set(a);
		for (int e=0; e<models.length; e++)
			models[e].setParam(a);
	}

	/**
	 *	Evaluate the total squared difference at the specified parameter
	 *	values.
	 *	@return					The squred difference
	 */
	public double f()
	{
		double result = 0;
		for (int e=0; e<models.length; e++)
		{
			double[] values = models[e].f();
			for (int i=0; i<values.length; i++)
				result += Math.pow(values[i] - yj[e][i*2+2], 2);
		}
		return result;
	}

	/**
	 *	Evaluate the gradient of the function at the specified parameter
	 *	values.
	 *	@return					The gradient
	 */
	public double[] gf()
	{
		double[] result = new double[xStored.length];
		for (int e=0; e<models.length; e++)
		{
			double[] values = models[e].f();
			double[][] gradients = models[e].gf();

			for (int k=0; k<xStored.length; k++)
				for (int i=0; i<values.length; i++)
					result[k] += 2 * (values[i] - yj[e][i*2+2]) * gradients[i][k];
		}
		return result;
	}

	/**
	 *	Evaluate the Hessian of the function at the specified parameter
	 *	values.
	 *	@return					The Hessian
	 */
	public double[][] hf()
	{
		double[][] result = new double[xStored.length][];
		for (int k=0; k<xStored.length; k++)
			result[k] = new double[xStored.length];

		for (int e=0; e<models.length; e++)
		{
			double[] values = models[e].f();
			double[][] gradients = models[e].gf();
			double[][][] hessians = models[e].hf();

			for (int k1=0; k1<xStored.length; k1++)
				for (int k2=0; k2<xStored.length; k2++)
					for (int i=0; i<values.length; i++)
						result[k1][k2] += 2 * (values[i] - yj[e][i*2+2]) * hessians[i][k1][k2] +
						                  2 * gradients[i][k1] * gradients[i][k2];
		}
		return result;
	}
}