package org.gersteinlab.regulation.model;

/**
 *	<pre>
 *	Providing methods for a rate model based on sigmoidal differential
 *	equations.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (September 1, 2008)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class SigmoidalRateModel extends RateModel
{
	/*----------------------------------------------------------------------
	 Instance variables
	----------------------------------------------------------------------*/

	// Model data
	protected double[][] yk = null;		// Values of the potential regulators
						// First dimension: potential regulator
						// Second dimension: time point

	protected double[][] dykdt = null;	// The differentials of the potential regulators
						// First dimension: potential regulator
						// Second dimension: time point

	// Function values, gradients and Hessians of the current model at the
	// specified time points
	protected double[] yj = null;		// The estimated values of yj at the specified time points
	protected double[][] gyj = null;	// The estimated gradients of yj at the specified time points
	protected double[][][] hyj = null;	// The estimated Hessians of yj at the specified time points

	// Frequently used values
	protected double[] expTerms = null;		// The exponential terms that are repeatedly used
	protected double hha1a1d2mha1p1 = Double.NaN;	// h * h * a[1] * a[1] / 2 - h * a[1] + 1
	protected double hha0a1 = Double.NaN;		// h * h * a[0] * a[1]


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	/**
	 *	@see		org.gersteinlab.regulation.model.RateModel#RateModel(double[], double, double[][][], double[][][])
	 */
	public SigmoidalRateModel(double[] tIn, double yj0In, double[][][] yksIn, double[][][] dykdtsIn)
	{
		super(tIn, yj0In, yksIn, dykdtsIn);
		yk = yks[0];
		dykdt = dykdts[0];
	}

	/**
	 *	Set the parameter values.
	 *	. a[0]: b_{j1}, maximum rate of expression
	 *	. a[1]: b_{j2}, degradation rate
	 *	. a[3]: a_{j0}, constant term in the linear part
	 *	. a[4..]: a_{jk}, coefficients of the regulators in the lienar
	 *	  part (corresponding to yk[0..])
	 *	If the values in the array are changed, call setParam() again to
	 *	clear any cached intermediate results before calling the other methods.
	 *	@param		aIn			The parameter values
	 */
	public void setParam(double[] aIn)
	{
		super.setParam(aIn);
		yj = null;
		gyj = null;
		hyj = null;
		expTerms = null;
		hha1a1d2mha1p1 = h * h * a[1] * a[1] / 2 - h * a[1] + 1;
		hha0a1 = h * h * a[0] * a[1];
	}

	/**
	 *	Note: to avoid array copying, the internal cached result is
	 *	returned. Calling methods should not modify the contents, or
	 *	should call setParam() first before calling this method again.
	 *	@see		org.gersteinlab.regulation.model.RateModel#f()
	 */
	public double[] f()
	{
		if (yj == null)
		{
			expTerms = new double[t.length];
			for (int i=0; i<t.length; i++)
			{
				expTerms[i] = a[2];
				for (int k=0; k<yk.length; k++)
					expTerms[i] += a[k+3] * yk[k][i];
				expTerms[i] = Math.exp(expTerms[i]);
			}

			yj = new double[n];
			for (int i=2; i<t.length; i+=2)	// For every next time point
			{
				int index = i / 2 - 1;
				double yn = (index == 0) ?yj0 :yj[index-1];

				yj[index] =   hha1a1d2mha1p1 * yn
				            + h * a[0] / (1 + expTerms[i-1])
				            - hha0a1 / 2 / (1 + expTerms[i-2]);
			}
		}

		return yj;
	}

	/**
	 *	Note: to avoid array copying, the internal cached result is
	 *	returned. Calling methods should not modify the contents, or
	 *	should call setParam() first before calling this method again.
	 *	@see		org.gersteinlab.regulation.model.RateModel#gf()
	 */
	public double[][] gf()
	{
		if (gyj == null)
		{
			if (yj == null)
				f();

			gyj = new double[n][];
			for (int i=2; i<t.length; i+=2)
			{
				int index = i / 2 - 1;
				gyj[index] = new double[a.length];
				double yn = (index == 0) ?yj0 :yj[index-1];

				gyj[index][0] =   hha1a1d2mha1p1 * (index==0 ?0 :gyj[index-1][0])
				                + h / (1 + expTerms[i-1])
				                - h * h * a[1] / 2 / (1 + expTerms[i-2]);

				gyj[index][1] =   h * (h * a[1] - 1) * yn
				                + hha1a1d2mha1p1 * (index==0 ?0 :gyj[index-1][1])
				                - h * h * a[0] / 2 / (1 + expTerms[i-2]);

				gyj[index][2] =   hha1a1d2mha1p1 * (index==0 ?0 :gyj[index-1][2])
				                - h * a[0] * expTerms[i-1] / Math.pow(1 + expTerms[i-1], 2)
				                + hha0a1 * expTerms[i-2] / 2 / Math.pow(1 + expTerms[i-2], 2);

				for (int k=0; k<yk.length; k++)
					gyj[index][k+3] =   hha1a1d2mha1p1 * (index==0 ?0 :gyj[index-1][k+3])
					                  - h * a[0] * yk[k][i-1] * expTerms[i-1] / Math.pow(1 + expTerms[i-1], 2)
					                  + hha0a1 * yk[k][i-2] * expTerms[i-2] / 2 / Math.pow(1 + expTerms[i-2], 2);
			}
		}
		return gyj;
	}

	/**
	 *	Note: to avoid array copying, the internal cached result is
	 *	returned. Calling methods should not modify the contents, or
	 *	should call setParam() first before calling this method again.
	 *	@see		org.gersteinlab.regulation.model.RateModel#hf()
	 */
	public double[][][] hf()
	{
		if (hyj == null)
		{
			if (yj == null)
				f();

			if (gyj == null)
				gf();

			hyj = new double[n][][];
			for (int i=2; i<t.length; i+=2)
			{
				int index = i / 2 - 1;
				hyj[index] = new double[a.length][];
				for (int k=0; k<a.length; k++)
					hyj[index][k] = new double[a.length];
				double yn = (index == 0) ?yj0 :yj[index-1];

				for (int k1=0; k1<a.length; k1++)
				{
					for (int k2=0; k2<a.length; k2++)
					{
						if (k1 == 0 || k2 == 0)
						{
							if (k1 == 0 && k2 == 0)
								hyj[index][k1][k2] =   0;
							else if (k1 == 1 || k2 == 1)
								hyj[index][k1][k2] =   h * (h * a[1] - 1) * (index==0 ?0 :gyj[index-1][0])
								                     + hha1a1d2mha1p1 * (index==0 ?0 :hyj[index-1][0][1])
								                     - h * h / 2 / (1 + expTerms[i-2]);
							else if (k1 == 2 || k2 == 2)
								hyj[index][k1][k2] =   hha1a1d2mha1p1 * (index==0 ?0 :hyj[index-1][0][2])
								                     - h * expTerms[i-1] / Math.pow(1 + expTerms[i-1], 2)
								                     + h * h * a[1] * expTerms[i-2] / 2 / Math.pow(1 + expTerms[i-2], 2);
							else
							{
								int k = (k1 == 0) ?(k2-3) :(k1-3);
								hyj[index][k1][k2] =   hha1a1d2mha1p1 * (index==0 ?0 :hyj[index-1][0][k+3])
								                     - h * yk[k][i-1] * expTerms[i-1] / Math.pow(1 + expTerms[i-1], 2)
								                     + h * h * a[1] * yk[k][i-2] * expTerms[i-2] / 2 / Math.pow(1 + expTerms[i-2], 2);
							}
						}
						else if (k1 == 1 || k2 == 1)
						{
							if (k1 == 1 && k2 == 1)
								hyj[index][k1][k2] =   h * h * yn
								                     + 2 * h * (h * a[1] - 1) * (index==0 ?0 :gyj[index-1][1])
								                     + hha1a1d2mha1p1 * (index==0 ?0 :hyj[index-1][1][1]);
							else if (k1 == 2 || k2 == 2)
								hyj[index][k1][k2] =   h * (h * a[1] - 1) * (index==0 ?0 :gyj[index-1][2])
								                     + hha1a1d2mha1p1 * (index==0 ?0 :hyj[index-1][1][2])
								                     + h * h * a[0] * expTerms[i-2] / 2 / Math.pow(1 + expTerms[i-2], 2);
							else
							{
								int k = (k1 == 1) ?(k2-3) :(k1-3);
								hyj[index][k1][k2] =   h * (h * a[1] - 1) * (index==0 ?0 :gyj[index-1][k+3])
								                     + hha1a1d2mha1p1 * (index==0 ?0 :hyj[index-1][1][k+3])
								                     + h * h * a[0] * yk[k][i-2] * expTerms[i-2] / 2 / Math.pow(1 + expTerms[i-2], 2);
							}
						}
						else if (k1 == 2 || k2 == 2)
						{
							if (k1 == 2 && k2 == 2)
								hyj[index][k1][k2] =   hha1a1d2mha1p1 * (index==0 ?0 :hyj[index-1][2][2])
								                     + h * a[0] * expTerms[i-1] * (expTerms[i-1] - 1) / Math.pow(1 + expTerms[i-1], 3)
								                     - hha0a1 * expTerms[i-2] * (expTerms[i-2] - 1) / 2 / Math.pow(1 + expTerms[i-2], 3);
							else
							{
								int k = (k1 == 2) ?(k2-3) :(k1-3);
								hyj[index][k1][k2] =   hha1a1d2mha1p1 * (index==0 ?0 :hyj[index-1][k+3][k+3])
								                     + h * a[0] * yk[k][i-1] * expTerms[i-1] * (expTerms[i-1] - 1) / Math.pow(1 + expTerms[i-1], 3)
								                     - hha0a1 * yk[k][i-2] * expTerms[i-2] * (expTerms[i-2] - 1) / 2 / Math.pow(1 + expTerms[i-2], 3);
							}
						}
						else
							hyj[index][k1][k2] =   hha1a1d2mha1p1 * (index==0 ?0 :hyj[index-1][k1][k2])
							                     + h * a[0] * yk[k1-3][i-1] * yk[k2-3][i-1] * expTerms[i-1] * (expTerms[i-1] - 1) / Math.pow(1 + expTerms[i-1], 3)
							                     - hha0a1 * yk[k1-3][i-2] * yk[k2-3][i-2] * expTerms[i-2] * (expTerms[i-2] - 1) / 2 / Math.pow(1 + expTerms[i-2], 3);
					}
				}
			}
		}
		return hyj;
	}
}