package org.gersteinlab.regulation.numeric;

/**
 *	<pre>
 *	A class for minimizing a differentiable multivariate real function using
 *	a descent method.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (September 4, 2008)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public abstract class DescentMethod
{
	/*----------------------------------------------------------------------
	 Constants
	----------------------------------------------------------------------*/

	public static final double ALPHA = 0.4;		// Stopping parameter in backtracking line search
	public static final double BETA = 0.2;		// Reduction factor of t in backtracking line search


	/*----------------------------------------------------------------------
	 Instance variables
	----------------------------------------------------------------------*/

	DifferentiableMultivariateFunction func = null;	// The function to be minimized


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	/**
	 *	Perform the minimization using default parameters.
	 *	@param		x0			The initial estimate of x
	 *	@return					The final estimate of x
	 */
	public abstract double[] minimize(double[] x0);


	/*----------------------------------------------------------------------
	 Non-public methods
	----------------------------------------------------------------------*/

	/**
	 *	The backtracking line search method (Armijo rule) for
	 *	determining step size at a point towards a descent direction.
	 *	@param		x			The point
	 *	@param		dx			The descent direction
	 *	@param		alpha			The stopping parameter
	 *	@param		beta			The reduction factor of t of backtracking line search
	 *	@return					The step size
	 */
	protected double armijo(double[] x, double[] dx, double alpha, double beta)
	{
		func.set(x);
		return armijo(x, func.f(), func.gf(), dx, alpha, beta);
	}

	/**
	 *	The backtracking line search method (Armijo rule) for
	 *	determining step size at a point towards a descent direction,
	 *	with precomputed function value and gradient.
	 *	@param		x			The point
	 *	@param		fx			The function value at the point
	 *	@param		gfx			The gradient at the point
	 *	@param		dx			The descent direction
	 *	@param		alpha			The stopping parameter
	 *	@param		beta			The reduction factor of t of backtracking line search
	 *	@return					The step size
	 */
	protected double armijo(double[] x, double fx, double[] gfx, double[] dx, double alpha, double beta)
	{
		int n = x.length;
		double t = 1;
		double gfxTdx = 0;
		for (int i=0; i<n; i++)
			gfxTdx += gfx[i] * dx[i];

		boolean done = false;
		while (!done)
		{
			double[] xPtd = new double[n];
			for (int i=0; i<n; i++)
				xPtd[i] = x[i] + t * dx[i];
			double fxPtd = func.f(xPtd);
			if (fxPtd > fx + alpha * t * gfxTdx)
				t *= beta;
			else
				done = true;
		}
		return t;
	}
}