package org.gersteinlab.regulation.numeric;

/**
 *	<pre>
 *	A class for minimizing a differentiable multivariate real function using
 *	gradient descent.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (September 4, 2008)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class GradientDescent extends DescentMethod
{
	/*----------------------------------------------------------------------
	 Constants
	----------------------------------------------------------------------*/

	public static final double ETA = 1e-6;		// Stopping parameter in gradient descent


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	/**
	 *	Construct an instance for a function.
	 *	@param		funcIn			The function
	 */
	public GradientDescent(DifferentiableMultivariateFunction funcIn)
	{
		func = funcIn;
	}

	/**
	 *	Perform the minimization using default parameters.
	 *	@param		x0			The initial estimate of x
	 *	@return					The final estimate of x
	 */
	public double[] minimize(double[] x0)
	{
		return minimize(x0, ALPHA, BETA, ETA);
	}

	/**
	 *	Perform the minimization.
	 *	@param		x0			The initial estimate of x
	 *	@param		alpha			The stopping parameter of backtracking line search
	 *	@param		beta			The reduction factor of t of backtracking line search
	 *	@param		eta			The stopping parameter
	 *	@return					The final estimate of x
	 */
	public double[] minimize(double[] x0, double alpha, double beta, double eta)
	{
		int n = x0.length;
		double[] x = new double[n];
		for (int i=0; i<n; i++)
			x[i] = x0[i];

		boolean done = false;
		while (!done)
		{
			func.set(x);
			double fx = func.f();
			double[] gfx = func.gf();
			double[] dx = new double[gfx.length];
			for (int i=0; i<gfx.length; i++)
				dx[i] = -1 * gfx[i];
			double norm = 0;
			for (int i=0; i<gfx.length; i++)
				norm += gfx[i] * gfx[i];
			norm = Math.pow(norm, 0.5);

			if (norm > eta)
			{
				double t = armijo(x, fx, gfx, dx, alpha, beta);
				boolean noChange = true;
				for (int i=0; i<n; i++)
				{
					if (x[i] + t * dx[i] != x[i])
						noChange = false;
					x[i] = x[i] + t * dx[i];
				}
				if (noChange)			// To avoid infinite loop when the step is too small
					done = true;
			}
			else
				done = true;
		}
		return(x);
	}
}