package org.gersteinlab.regulation.numeric;

import Jama.*;

/**
 *	<pre>
 *	A class for minimizing a doubly differentiable multivariate real
 *	function using the Newton's method.
 *	</pre>
 *
 *	@author		Kevin Yuk-Lap Yip
 *	@version	1.0 (August 29, 2008)
 *
 *	<pre>
 *	Change History:
 *	1.0	- Initial version
 *	</pre>
 */
public class NewtonMethod extends DescentMethod
{
	/*----------------------------------------------------------------------
	 Constants
	----------------------------------------------------------------------*/

	public static final double EPSILON = 1e-6;	// Stopping parameter


	/*----------------------------------------------------------------------
	 Public methods
	----------------------------------------------------------------------*/

	/**
	 *	Construct an instance for a function.
	 *	@param		funcIn			The function
	 */
	public NewtonMethod(DoublyDifferentiableMultivariateFunction funcIn)
	{
		func = funcIn;
	}

	/**
	 *	Perform the minimization using default parameters.
	 *	@param		x0			The initial estimate of x
	 *	@return					The final estimate of x
	 */
	public double[] minimize(double[] x0)
	{
		return minimize(x0, ALPHA, BETA, EPSILON);
	}

	/**
	 *	Perform the minimization.
	 *	@param		x0			The initial estimate of x
	 *	@param		alpha			The stopping parameter of backtracking line search
	 *	@param		beta			The reduction factor of t of backtracking line search
	 *	@param		epsilon			The stopping parameter
	 *	@return					The final estimate of x
	 */
	public double[] minimize(double[] x0, double alpha, double beta, double epsilon)
	{
		int n = x0.length;
		double[] x = new double[n];
		for (int i=0; i<n; i++)
			x[i] = x0[i];

		boolean done = false;
		int iter = 0;
		while (!done)
		{
			func.set(x);
			double fx = func.f();
			double[] gfx = func.gf();
			double[][] hfx = ((DoublyDifferentiableMultivariateFunction)func).hf();

			double[] dx = nd(gfx, hfx);
			double lambda2x = lambda2(dx, hfx);

			if (lambda2x / 2 > epsilon)
			{
				double t = armijo(x, fx, gfx, dx, alpha, beta);
				boolean noChange = true;
				for (int i=0; i<n; i++)
				{
					if (x[i] + t * dx[i] != x[i])
						noChange = false;
					x[i] = x[i] + t * dx[i];
				}
				if (noChange)			// To avoid infinite loop when the step is too small
					done = true;
			}
			else
				done = true;
			iter ++;
			if (iter == 200)
				done = true;
		}
		return(x);
	}


	/*----------------------------------------------------------------------
	 Non-public methods
	----------------------------------------------------------------------*/

	/**
	 *	Get the standard Newton's direction at a point.
	 *	@param		x			The point
	 *	@return					The Newton's direction at the point
	 */
	protected double[] nd(double[] x)
	{
		func.set(x);
		return nd(func.gf(), ((DoublyDifferentiableMultivariateFunction)func).hf());
	}

	/**
	 *	Get the standard Newton's direction at a point, with precomputed
	 *	gradient and Hessian.
	 *	@param		gfx			The gradient at the point
	 *	@param		hfx			The Hessian at the point
	 *	@return					The Newton's direction at the point
	 */
	protected double[] nd(double[] gfx, double[][] hfx)
	{
		int n = gfx.length;
		Matrix hfxm = new Matrix(hfx, n, n);
		Matrix hfxm_1 = hfxm.inverse();
		Matrix gfxm = new Matrix(gfx, n);
		Matrix _gfxm = gfxm.timesEquals(-1);
		return hfxm_1.times(_gfxm).getColumnPackedCopy();
	}

	/**
	 *	Get the error estimate, lambda^2, at a point.
	 *	@param		x			The point
	 *	@return					lambda^2 at the point
	 */
	protected double lambda2(double[] x)
	{
		func.set(x);
		return lambda2(nd(x), ((DoublyDifferentiableMultivariateFunction)func).hf());
	}

	/**
	 *	Get the error estimate, lambda^2, at a point, with precomputed
	 *	Newton's direction and Hessian.
	 *	@param		dx			The Newton's direction at the point
	 *	@param		hfx			The Hessian at the point
	 *	@return					lambda^2 at the point
	 */
	protected double lambda2(double[] dx, double[][] hfx)
	{
		int n = dx.length;
		Matrix dxm = new Matrix(dx, n);
		Matrix hfxm = new Matrix(hfx, n, n);
		return dxm.transpose().times(hfxm).times(dxm).get(0, 0);
	}
}